import click

from sagasu import util as u
from sagasu.config import ConfigUtil
from sagasu.crawler import CrawlerEngine
from sagasu.engine import SearchEngine


@click.command()
@click.argument('mode')
def app(mode):
  if mode not in ["search", "crawl", "indexing"]:
    print("choice in [search, crawl, indexing]")
    return

  if not u.is_exist("/config/config.yml"):
    print("please set ~/.sagasu/config/config.yml")
    return

  config = ConfigUtil().load()
  engine = SearchEngine(config)

  if mode == "indexing":
    crawler_engine = CrawlerEngine(config.sources)
    crawler_engine.crawl_all()
    engine.indexing()
  elif mode == "search":
    word = input("let's type search word >>> ")
    resources = engine.word_search(word)
    for resource in resources:
      print(f"""
  [URI]
    {resource.uri}
  [Sentence]
    {resource.sentence[:30]}...
      """)


if __name__ == '__main__':
  app()



from dataclasses import dataclass
from typing import List

import yaml

from sagasu import util


@dataclass
class SourceModel:
  source_type: str
  target: str


@dataclass
class ConfigModel:
  sources: List[SourceModel]


class ConfigUtil:
  def __init__(self):
    self.path = f"{util.SAGASU_WORKDIR}/config/config.yml"

  def load(self) -> ConfigModel:
    with open(self.path) as f:
      _config = yaml.safe_load(f)
      config = ConfigModel(
        sources=[
          SourceModel(
            source_type=c['source_type'],
            target=c['target']
          )
          for c in _config['sources']
        ]
      )
    return config



import tensorflow as tf


class BahdanauAttention(tf.keras.Model):
  def __init__(self, units):
    super(BahdanauAttention, self).__init__()
    self.W1 = tf.keras.layers.Dense(units)
    self.W2 = tf.keras.layers.Dense(units)
    self.V = tf.keras.layers.Dense(1)

  def call(self, features, hidden):
    hidden_with_time_axis = tf.expand_dims(hidden, 1)
    score = tf.nn.tanh(self.W1(features) + self.W2(hidden_with_time_axis))
    attention_weights = tf.nn.softmax(self.V(score), axis=1)
    context_vector = attention_weights * features
    context_vector = tf.reduce_sum(context_vector, axis=1)

    return context_vector, attention_weights


class CNNEncoder(tf.keras.Model):
  def __init__(self, embedding_dim):
    super(CNNEncoder, self).__init__()
    self.fc = tf.keras.layers.Dense(embedding_dim)

  def call(self, x):
    x = self.fc(x)
    x = tf.nn.relu(x)
    return x


class RNNDecoder(tf.keras.Model):
  def __init__(self, embedding_dim, units, vocab_size):
    super(RNNDecoder, self).__init__()
    self.units = units

    self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim)
    self.gru = tf.keras.layers.GRU(self.units,
                                   return_sequences=True,
                                   return_state=True,
                                   recurrent_initializer='glorot_uniform')
    self.fc1 = tf.keras.layers.Dense(self.units)
    self.fc2 = tf.keras.layers.Dense(vocab_size)

    self.attention = BahdanauAttention(self.units)

  def call(self, x, features, hidden):
    context_vector, attention_weights = self.attention(features, hidden)
    x = self.embedding(x)
    x = tf.concat([tf.expand_dims(context_vector, 1), x], axis=-1)
    output, state = self.gru(x)
    x = self.fc1(output)
    x = tf.reshape(x, (-1, x.shape[2]))
    x = self.fc2(x)
    return x, state, attention_weights

  def reset_state(self, batch_size):
    return tf.zeros((batch_size, self.units))


if __name__ == '__main__':
  images = tf.random.normal([1, 224, 224, 3])

  image_model = tf.keras.applications.MobileNetV2(
    include_top=False, weights='imagenet')
  new_input = image_model.input
  hidden_layer = image_model.layers[-1].output

  image_features_extract_model = tf.keras.Model(new_input, hidden_layer)
  features = image_features_extract_model(images)
  features = tf.reshape(features, (features.shape[0], -1, features.shape[3]))

  enc = CNNEncoder(embedding_dim=256)
  dec = RNNDecoder(embedding_dim=256, units=512, vocab_size=9880)

  dec.load_weights('../model/dec/dec_save_weights')

  enc.load_weights("../model/enc/enc_save_weights")
  encoded = enc(features)
  hidden = dec.reset_state(batch_size=1)
  print(encoded.shape)

  with open('../model/tokenize/token.json') as f:
    s = f.readline()
  tokenizer = tf.keras.preprocessing.text.tokenizer_from_json(s)
  print((iw := tokenizer.index_word)[0], iw[1], iw[2], iw[3])

  result = []
  dec_input = tf.expand_dims([tokenizer.word_index['<start>']], 0)

  for i in range(100):
    print(result)
    print(dec_input.shape, features.shape, hidden.shape)
    predictions, hidden, attention_weights = dec(dec_input, encoded, hidden)

    predicted_id = tf.random.categorical(predictions, 1)[0][0].numpy()
    result.append(tokenizer.index_word[predicted_id])

    if tokenizer.index_word[predicted_id] == '<end>':
      break
    else:
      dec_input = tf.expand_dims([predicted_id], 0)



from dataclasses import dataclass
from typing import List, TypedDict, Union


@dataclass
class Resource:
  uri: str
  sentence: str


@dataclass
class TokenizedResource:
  index: List[str]
  resource: Resource


@dataclass
class IndexedResource(TypedDict):
  index: str
  resource: List[Resource]


@dataclass
class SampleResource(Resource):
  pass


@dataclass
class TwitterResource(Resource):
  image_urls: Union[List[str], None]
  image_captions: Union[List[str], None]


@dataclass
class ScrapboxResource(Resource):
  image_urls: Union[List[str], None]
  image_captions: Union[List[str], None]


@dataclass
class SlackResource(Resource):
  pass



import os
from pathlib import Path
from typing import List

import pandas as pd

from sagasu import model as m
from sagasu import util as u
from sagasu import crawler as c


class Repository:
  def load(self) -> List[m.Resource]:
    raise NotImplementedError("")


class SampleRepository(Repository):
  def __init__(self):
    u.mkdir_p((p := f'{u.SAGASU_WORKDIR}/sample/sample.txt'))
    with open(p, 'w') as f:
      f.write("""
吾輩は猫である。
親譲りの無鉄砲で小供の時から損ばかりしている。
""")

  def load(self) -> List[m.SampleResource]:
    with open(p := f'{u.SAGASU_WORKDIR}/sample/sample.txt') as f:
      document = f.readlines()
      document = ''.join(document)
    return [m.SampleResource(uri=p, sentence=document)]


class TwitterRepository(Repository):

  def load(self) -> List[m.TwitterResource]:
    dir_path = Path(f"{c.CRAWLER_WORK_DIR}/twitter/uri-sentence")
    uri_sentence_dfs = []
    for file in dir_path.iterdir():
      if not file.is_file():
        continue
      df = pd.read_csv(file, sep='\t')
      uri_sentence_dfs.append(df)
    uri_sentence_df = pd.concat(uri_sentence_dfs).reset_index()

    dir_path = Path(f"{c.CRAWLER_WORK_DIR}/twitter/uri-media")
    uri_media_dfs = []
    for file in dir_path.iterdir():
      if not file.is_file():
        continue
      df = pd.read_csv(file, sep='\t')
      uri_media_dfs.append(df)
    uri_media_df = pd.concat(uri_media_dfs).reset_index()

    df = pd.merge(uri_sentence_df, uri_media_df)
    # media_url1  media_url2  media_url3  media_url4  media_caption1
    resources = [m.TwitterResource(
      uri=r[1].uri, sentence=r[1].sentence,
      image_urls=[r[1].media_url1, r[1].media_url2, r[1].media_url3, r[1].media_url4],
      image_captions=[r[1].media_caption1, r[1].media_caption2, r[1].media_caption3, r[1].media_caption4]
    ) for r in df.iterrows()]

    return resources


class ScrapboxRepository(Repository):
  def load(self) -> List[m.ScrapboxResource]:
    dir_path = Path(f"{c.CRAWLER_WORK_DIR}/scrapbox/uri-sentence")
    uri_sentence_dfs = []
    for file in dir_path.iterdir():
      if not file.is_file():
        continue
      df = pd.read_csv(file, sep='\t')
      uri_sentence_dfs.append(df)
    uri_sentence_df = pd.concat(uri_sentence_dfs).reset_index()

    dir_path = Path(f"{c.CRAWLER_WORK_DIR}/scrapbox/uri-media")
    uri_media_dfs = []
    for file in dir_path.iterdir():
      if not file.is_file():
        continue
      df = pd.read_csv(file, sep='\t')
      uri_media_dfs.append(df)
    uri_media_df = pd.concat(uri_media_dfs).reset_index()

    df = pd.merge(uri_sentence_df, uri_media_df)
    # media_url1  media_url2  media_url3  media_url4  media_caption1
    resources = [m.ScrapboxResource(
      uri=r[1].uri, sentence=r[1].sentence,
      image_urls=[r[1].media_url1, r[1].media_url2, r[1].media_url3, r[1].media_url4],
      image_captions=[r[1].media_caption1, r[1].media_caption2, r[1].media_caption3, r[1].media_caption4]
    ) for r in df.iterrows()]

    return resources



from typing import List
import os
import pickle

import spacy
import datetime

from sagasu.model import Resource, IndexedResource
from sagasu.util import mkdir_p

nlp = spacy.load('ja_ginza')
JST = datetime.timezone(datetime.timedelta(hours=+9), 'JST')


def simple_n_gram(sentence: str, n: int) -> List[str]:
  index = [sentence[idx:idx + n] for idx in range(len(sentence) - n + 1)]
  return index


def word_n_gram(sentence: str, n: int = 1) -> List[str]:
  doc = nlp(sentence)
  tokens = []
  for sent in doc.sents:
    for token in sent:
      tokens.append(token.orth_)
  index = [''.join(tokens[idx:idx + n]) for idx in range(len(tokens) - n + 1)]
  return index


def dump(func):
  def runner(*args, **kwargs):
    result = func(*args, **kwargs)
    t = datetime.datetime.now(JST)
    path_prefix = f"{os.getenv('HOME')}/.sagasu"
    filename_prefix = f"/{t.year}-" \
                      f"{m if len(m := str(t.month)) == 2 else f'0{m}'}-" \
                      f"{d if len(d := str(t.day)) == 2 else f'0{d}'}-" \
                      f"{h if len(h := str(t.hour)) == 2 else f'{h}'}"

    mkdir_p(file_path := path_prefix + "/indexed" + filename_prefix + ".pkl")
    with open(file_path, 'wb') as f:
      pickle.dump(result, f)
    return result

  return runner


class Indexer:
  def __call__(self, resource: Resource, indexed: IndexedResource) -> IndexedResource:
    raise NotImplementedError("not implemented")


class WordUnigramIndexer(Indexer):
  def __call__(self, resource: Resource, indexed: IndexedResource) -> IndexedResource:
    indices = word_n_gram(resource.sentence)
    for index in indices:
      if index not in indexed:
        indexed[index] = [resource]
      else:
        indexed[index].append(resource)
    return indexed


class WordNgramIndexer(Indexer):
  def __init__(self, n):
    self.n = n

  @dump
  def __call__(self, resource: Resource, indexed: IndexedResource) -> IndexedResource:
    indices = word_n_gram(resource.sentence, self.n)
    for index in indices:
      if index not in indexed:
        indexed[index] = [resource]
      else:
        indexed[index].append(resource)
    return indexed


if __name__ == '__main__':
  print(simple_n_gram('毎日の料理を楽しみにする', 2))
  print(word_n_gram('毎日の料理を楽しみにする'))
  print(word_n_gram('毎日の料理を楽しみにする', 2))
